from .tokenizer import tokens
from .tokenizer.tokens import OperatorType
from ..token_scanner import TokenScanner

class FunctionCall:
    __slots__ = ('type', 'name', 'arguments')

    def __init__(self, name: str, arguments: list):
        self.type = 'FunctionCall'
        self.name = name
        self.arguments = arguments

class CSSValue:
    __slots__ = ('type', 'value')

    def __init__(self, value: list):
        self.type = 'CSSValue'
        self.value = value


class CSSProperty:
    __slots__ = ('name', 'value', 'important', 'snippet')

    def __init__(self, name: str, value: list, important=False):
        self.name = name
        self.value = value
        self.important = important
        self.snippet = None
        "Snippet matched with current property"

def parser(token_list: list, options={}):
    scanner = TokenScanner(token_list)
    result = []

    while scanner.readable():
        prop = consume_property(scanner, options)
        if prop:
            result.append(prop)
        elif not scanner.consume(is_sibling_operator):
            raise scanner.error('Unexpected token')

    return result

def consume_property(scanner: TokenScanner, options={}):
    "Consumes single CSS property"
    name = None
    important = False
    value_fragment = None
    value = []
    token = scanner.peek()
    value_mode = bool(options.get('value', False))

    if not value_mode and is_literal(token) and not is_function_start(scanner):
        scanner.pos += 1
        name = token.value
        # Consume any following value delimiter after property name
        scanner.consume(is_value_delimiter)

    # Skip whitespace right after property name, if any
    if value_mode:
        scanner.consume(is_white_space)

    while scanner.readable():
        if scanner.consume(is_important):
            important = True
        else:
            value_fragment = consume_value(scanner, value_mode)
            if value_fragment:
                value.append(value_fragment)
            elif not scanner.consume(is_fragment_delimiter):
                break

    if name or value or important:
        return CSSProperty(name, value, important)

def consume_value(scanner: TokenScanner, in_argument=False):
    "Consumes single value fragment, e.g. all value tokens before comma"
    result = []
    token = None
    args = None

    while scanner.readable():
        token = scanner.peek()
        if is_value(token):
            scanner.pos += 1
            args = consume_arguments(scanner) if is_literal(token) else None

            if args is not None:
                result.append(FunctionCall(token.value, args))
            else:
                result.append(token)
        elif is_value_delimiter(token) or (in_argument and is_white_space(token)):
            scanner.pos += 1
        else:
            break

    return CSSValue(result) if result else None

def consume_arguments(scanner: TokenScanner):
    start = scanner.pos
    if scanner.consume(is_open_bracket):
        args = []
        value = None

        while scanner.readable() and not scanner.consume(is_close_bracket):
            value = consume_value(scanner, True)
            if value:
                args.append(value)
            elif not scanner.consume(is_white_space) and not scanner.consume(is_argument_delimiter):
                raise scanner.error('Unexpected token')

        scanner.start = start
        return args

def is_literal(token: tokens.Token):
    return isinstance(token, tokens.Literal)


def is_bracket(token: tokens.Token, is_open=None):
    return isinstance(token, tokens.Bracket) and (is_open is None or token.open == is_open)


def is_open_bracket(token: tokens.Token):
    return is_bracket(token, True)


def is_close_bracket(token: tokens.Token):
    return is_bracket(token, False)


def is_white_space(token: tokens.Token):
    return isinstance(token, tokens.WhiteSpace)


def is_operator(token: tokens.Token, operator: OperatorType=None):
    return isinstance(token, tokens.Operator) and (not operator or token.operator == operator)


def is_sibling_operator(token: tokens.Token):
    return is_operator(token, OperatorType.Sibling)


def is_argument_delimiter(token: tokens.Token):
    return is_operator(token, OperatorType.ArgumentDelimiter)


def is_fragment_delimiter(token: tokens.Token):
    return is_argument_delimiter(token)

def is_important(token: tokens.Token):
    return is_operator(token, OperatorType.Important)


def is_value(token: tokens.Token):
    return isinstance(token, tokens.StringValue) or \
        isinstance(token, tokens.ColorValue) or \
        isinstance(token, tokens.NumberValue) or \
        isinstance(token, tokens.Literal) or \
        isinstance(token, tokens.Field) or \
        isinstance(token, tokens.CustomProperty)


def is_value_delimiter(token: tokens.Token):
    return is_operator(token, OperatorType.PropertyDelimiter) or \
        is_operator(token, OperatorType.ValueDelimiter)


def is_function_start(scanner: TokenScanner):
    max_ix = len(scanner.tokens) - 1
    if scanner.pos < max_ix:
        t1 = scanner.tokens[scanner.pos]
        t2 = scanner.tokens[scanner.pos + 1]
        return t1 and t2 and is_literal(t1) and isinstance(t2, tokens.Bracket)



from ...abbreviation import AbbreviationNode, AbbreviationAttribute
from ..utils import find

def label(node: AbbreviationNode):
    if node.name == 'label':
        input = find(node, lambda n: n.name == 'input' or n.name == 'textarea')
        if input:
            # Remove empty `for` attribute
            if node.attributes:
                node.attributes = [attr for attr in node.attributes if not (attr.name == 'for' and is_empty_attribute(attr))]

            # Remove empty `id` attribute
            if input.attributes:
                input.attributes = [attr for attr in input.attributes if not (attr.name == 'id' and is_empty_attribute(attr))]


def is_empty_attribute(attr: AbbreviationAttribute):
    if not attr.value:
        return True


    if len(attr.value) == 1:
        token = attr.value[0]
        if token and not isinstance(token, str) and not token.name:
            # Attribute contains field
            return True

    return False





from .tokenizer import tokens


operators = {
    'child': '>',
    'class': '.',
    'climb': '^',
    'id': '#',
    'equal': '=',
    'close': '/',
    'sibling': '+'
}


def stringify(token: tokens.Token, state):
    visitor = globals().get(token.type)
    if not visitor:
        raise Exception('Unknown token %s' % token.type)

    return visitor(token, state)


def Literal(token: tokens.Literal, state):
    return token.value


def Quote(token: tokens.Quote, state):
    return '\'' if token.single else '"'


def Bracket(token: tokens.Bracket, state):
    if token.context == 'attribute':
        return '[' if token.open else ']'
    if token.context == 'expression':
        return '{' if token.open else '}'
    return '(' if token.open else '}'


def Operator(token: tokens.Operator, state):
    global operators
    return operators.get(token.operator)


def Field(token: tokens.Field, state):
    if token.index is not None:
        # It’s a field: by default, return TextMate-compatible field
        fmt = '${%d:%s}' if token.name else '${%s}'
        return fmt % (token.index, token.name)

    if token.name:
        # It’s a variable
        return state.get_variable(token.name)

    return ''


def RepeaterPlaceholder(token: tokens.RepeaterPlaceholder, state):
    # Find closest implicit repeater
    repeater = None
    repeater_list = state.repeaters[:]
    repeater_list.reverse()

    for r in repeater_list:
        if r.implicit:
            repeater = r
            break

    state.inserted = True
    return state.get_text(repeater.value) if repeater else None


def RepeaterNumber(token: tokens.RepeaterNumber, state):
    value = 1
    last_ix = len(state.repeaters) - 1

    if last_ix >= 0:
        repeater = state.repeaters[-1]
        if token.reverse:
            value = token.base + repeater.count - repeater.value - 1
        else:
            value = token.base + repeater.value

        if token.parent:
            parent_ix = max(0, last_ix - token.parent)
            if parent_ix != last_ix:
                value += repeater.count * state.repeaters[parent_ix].value

    result = str(value)
    prefix = '0' * max(0, token.size - len(result))

    return prefix + result


def WhiteSpace(token: tokens.WhiteSpace, state):
    return token.value



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="py-emmet",
    version="1.3.1",
    author="Sergey Chikuyonok",
    author_email="serge.che@gmail.com",
    description="Emmet is a web-developer’s toolkit for boosting HTML & CSS code writing.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/emmetio/py-emmet",
    include_package_data=True,
    packages=setuptools.find_packages(exclude=('tests', 'tests.*',)),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.3',
)



import unittest
import sys

sys.path.append('../')

from emmet.config import Config
from emmet.markup import parse
from .stringify import stringify_node

default_config = Config()
bem = Config({ 'options': { 'bem.enabled': True }, 'cache': {} })

def expand(abbr: str, config=default_config):
    return stringify_node(parse(abbr, config))


class TestMarkupAbbreviation(unittest.TestCase):
    def test_implicit_tags(self):
        self.assertEqual(expand('.'), '<div class=""></div>')
        self.assertEqual(expand('.foo>.bar'), '<div class="foo"><div class="bar"></div></div>')
        self.assertEqual(expand('p.foo>.bar'), '<p class="foo"><span class="bar"></span></p>')
        self.assertEqual(expand('ul>.item*2'), '<ul><li*2@0 class="item"></li><li*2@1 class="item"></li></ul>')
        self.assertEqual(expand('table>.row>.cell'), '<table><tr class="row"><td class="cell"></td></tr></table>')
        self.assertEqual(expand('{test}'), 'test')
        self.assertEqual(expand('.{test}'), '<div class="">test</div>')
        self.assertEqual(expand('ul>.item$*2'), '<ul><li*2@0 class="item1"></li><li*2@1 class="item2"></li></ul>')

    def test_xsl(self):
        config = Config({ 'syntax': 'xsl' })
        self.assertEqual(expand('xsl:variable[select]', config), '<xsl:variable select=""></xsl:variable>')
        self.assertEqual(expand('xsl:with-param[select]', config), '<xsl:with-param select=""></xsl:with-param>')
        self.assertEqual(expand('xsl:variable[select]>div', config), '<xsl:variable><div></div></xsl:variable>')
        self.assertEqual(expand('xsl:with-param[select]{foo}', config), '<xsl:with-param>foo</xsl:with-param>')

class TestLabelPreprocessor(unittest.TestCase):
    def test_label(self):
        self.assertEqual(expand('label>input'), '<label><input type="${1:text}" /></label>')
        self.assertEqual(expand('label>inp'), '<label><input type="${1:text}" name="${1}" /></label>')
        self.assertEqual(expand('label>span>input'), '<label><span><input type="${1:text}" /></span></label>')
        self.assertEqual(expand('label+inp'), '<label for=""></label><input type="${1:text}" name="${1}" id="${1}" />')

class TestBEMTransform(unittest.TestCase):
    def test_modifiers(self):
        self.assertEqual(expand('div.b_m', bem), '<div class="b b_m"></div>')
        self.assertEqual(expand('div.b._m', bem), '<div class="b b_m"></div>')
        self.assertEqual(expand('div.b_m1._m2', bem), '<div class="b b_m1 b_m2"></div>')
        self.assertEqual(expand('div.b>div._m', bem), '<div class="b"><div class="b b_m"></div></div>')
        self.assertEqual(expand('div.b>div._m1>div._m2', bem), '<div class="b"><div class="b b_m1"><div class="b b_m2"></div></div></div>')

        # classnames with -
        self.assertEqual(expand('div.b>div._m1-m2', bem), '<div class="b"><div class="b b_m1-m2"></div></div>')

    def test_elements(self):
        self.assertEqual(expand('div.b>div.-e', bem), '<div class="b"><div class="b__e"></div></div>')
        self.assertEqual(expand('div.b>div.---e', bem), '<div class="b"><div class="b__e"></div></div>')
        self.assertEqual(expand('div.b>div.-e>div.-e', bem), '<div class="b"><div class="b__e"><div class="b__e"></div></div></div>')
        self.assertEqual(expand('div', bem), '<div></div>', 'Fixes bug with empty class')

        # get block name from proper ancestor
        self.assertEqual(expand('div.b1>div.b2_m1>div.-e1+div.---e2_m2', bem),
            '<div class="b1"><div class="b2 b2_m1"><div class="b2__e1"></div><div class="b1__e2 b1__e2_m2"></div></div></div>')

        # class names with -
        self.assertEqual(expand('div.b>div.-m1-m2', bem), '<div class="b"><div class="b__m1-m2"></div></div>')

        # class names with _
        self.assertEqual(expand('div.b_m_o', bem), '<div class="b b_m_o"></div>')

    def test_customize_modifier(self):
        conf = Config({
            'options': {
                'bem.enabled': True,
                'bem.element': '-',
                'bem.modifier': '__'
            }
        })
        self.assertEqual(expand('div.b_m', conf), '<div class="b b__m"></div>')
        self.assertEqual(expand('div.b._m', conf), '<div class="b b__m"></div>')

    def test_multiple_classes(self):
        self.assertEqual(expand('div.b_m.c', bem), '<div class="b b_m c"></div>')
        self.assertEqual(expand('div.b>div._m.c', bem), '<div class="b"><div class="b b_m c"></div></div>')
        self.assertEqual(expand('div.b>div.-m.c', bem), '<div class="b"><div class="b__m c"></div></div>')

    def test_parent_context(self):
        conf = Config({
            'context': { 'name': 'div', 'attributes': { 'class': 'bl' } },
            'options': { 'bem.enabled': True }
        })
        self.assertEqual(expand('.-e_m', conf), '<div class="bl__e bl__e_m"></div>')



import re
from random import randint
from ...abbreviation import AbbreviationNode
from ...config import Config
from ..implicit_tag import resolve_implicit_tag
from .latin import vocabulary as latin
from .russian import vocabulary as ru
from .spanish import vocabulary as sp

re_lorem = re.compile(r'^lorem([a-z]*)(\d*)(-\d*)?$', re.I)

vocabularies = {
    'ru': ru,
    'sp': sp,
    'latin': latin
}

def lorem(node: AbbreviationNode, ancestors: list, config: Config):
    if not node.name:
        return

    m = re_lorem.match(node.name)
    if m:
        db = vocabularies.get(m.group(1)) or vocabularies.get('latin')
        min_word_count = max(1, int(m.group(2))) if m.group(2) else 30
        max_word_count = max(min_word_count, int(m.group(3)[1:])) if m.group(3) else min_word_count
        word_count = randint(min_word_count, max_word_count)
        repeat = node.repeat or find_repeater(ancestors)

        node.name = node.attributes = None
        node.value = [paragraph(db, word_count, not repeat or repeat.value == 0)]

        if node.repeat and len(ancestors) > 1:
            resolve_implicit_tag(node, ancestors, config)

def sample(arr: list, count: int):
    l = len(arr)
    iterations = min(l, count)
    result = []

    while len(result) < iterations:
        item = arr[randint(0, l - 1)]
        if item not in result:
            result.append(item)

    return result


def choice(val: str):
    return val[randint(0, len(val) - 1)]


def sentence(words: list, end: str=None):
    if words:
        words = [words[0].capitalize()] + words[1:]

    return ' '.join(words) + (end or choice('?!...')) # more dots than question marks

def insert_commas(words: list):
    """
    Insert commas at randomly selected words. This function modifies values
    inside `words` array
    """
    if len(words) < 2:
        return words

    words = words[:]
    l = len(words)
    total_commas = 0

    if 3 < l <= 6:
        total_commas = randint(0, 1)
    elif 6 < l <= 12:
        total_commas = randint(0, 2)
    else:
        total_commas = randint(1, 4)

    for _ in range(total_commas):
        pos = randint(0, l - 2)
        if words[pos][-1] != ',':
            words[pos] += ','

    return words

def paragraph(db: dict, word_count: int, start_with_common=False):
    """
    Generate a paragraph of "Lorem ipsum" text
    :param db Words dictionary
    :param word_count Words count in paragraph
    :param start_with_common Should paragraph start with common "lorem ipsum" sentence.
    """
    result = []
    total_words = 0

    if start_with_common and 'common' in db:
        words = db['common'][0:word_count]
        total_words += len(words)
        result.append(sentence(insert_commas(words), '.'))

    while total_words < word_count:
        words = sample(db['words'], min(randint(2, 30), word_count - total_words))
        total_words += len(words)
        result.append(sentence(insert_commas(words)))

    return ' '.join(result)


def find_repeater(ancestors: list):
    i = len(ancestors) - 1
    while i >= 0:
        element = ancestors[i]
        if isinstance(element, AbbreviationNode) and element.repeat:
            return element.repeat
        i -= 1



from .html import html
from .indent_format import indent_format
from ...abbreviation import Abbreviation
from ...config import Config


def haml(abbr: Abbreviation, config: Config):
    return indent_format(abbr, config, {
        'beforeName': '%',
        'beforeAttribute': '(',
        'afterAttribute': ')',
        'glueAttribute': ' ',
        'afterTextLine': ' |',
        'booleanValue': 'true',
        'selfClose': '/'
    })


def pug(abbr: Abbreviation, config: Config):
    return indent_format(abbr, config, {
        'beforeAttribute': '(',
        'afterAttribute': ')',
        'glueAttribute': ', ',
        'beforeTextLine': '| ',
        'selfClose': '/' if config.options.get('output.selfClosingStyle') == 'xml' else ''
    })


def slim(abbr: Abbreviation, config: Config):
    return indent_format(abbr, config, {
        'beforeAttribute': ' ',
        'glueAttribute': ' ',
        'beforeTextLine': '| ',
        'selfClose': '/'
    })




from ...abbreviation import AbbreviationNode, AbbreviationAttribute

def xsl(node: AbbreviationNode):
    "XSL transformer: removes `select` attributes from certain nodes that contain children"
    if matches_name(node.name) and node.attributes and (node.children or node.value):
        node.attributes = filter(is_allowed, node.attributes)

def is_allowed(attr: AbbreviationAttribute):
    return attr.name != 'select'

def matches_name(name: str):
    return name == 'xsl:variable' or name == 'xsl:with-param'




class CSSAbbreviationScope:
    Global = '@@global'
    "Include all possible snippets in match"

    Section = '@@section'
    "Include raw snippets only (e.g. no properties) in abbreviation match"

    Property = '@@property'
    "Include properties only in abbreviation match"

    Value = '@@value'
    "Resolve abbreviation in context of CSS property value"



import unittest
import sys

sys.path.append('../')

from emmet import expand
from emmet.config import markup_snippets, xsl_snippets
from emmet.abbreviation import parse


class TestSnippets(unittest.TestCase):
    def test_html(self):
        for _, v in enumerate(markup_snippets):
            self.assertTrue(parse(v))

        self.assertEqual(expand('fset>input:c'), '<fieldset><input type="checkbox" name="" id=""></fieldset>')

    def test_xsl(self):
        for _, v in enumerate(xsl_snippets):
            self.assertTrue(parse(v))
